if Ace3 then
  HazLoot = LibStub("AceAddon-3.0"):NewAddon("HazLoot", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0")
else -- not in WoW
  HazLoot = {}
  self = HazLoot
  HazLoot.standalone = true
  function HazLoot:Print(message)
    print("[Core] "..message)
  end
  function HazLoot:Debug(message)
    print("[Core][DEBUG] "..message)
  end
  require("HazLootList")
end

--TODO: Separate out recipes, patterns, gems, pets etc.
--TODO: Implement timers for bidding process
--TODO: Update loot lists when thresholds are changed.
--TODO: maybe rework the loot list to use periodic table?
--TODO: Present a frame for the suicide -> last call -> off-spec -> dust

-- Constants
HazLoot.TRASH     = 0
HazLoot.COMMON    = 1
HazLoot.UNCOMMON  = 2
HazLoot.RARE      = 3
HazLoot.EPIC      = 4
HazLoot.LEGENDARY = 5
HazLoot.ARTIFACT  = 6

HazLoot.LootTypes = {"Common","Uncommon","Rare","Epic","Legendary","Artifact" }
HazLoot.LootTypes[HazLoot.TRASH]     = "Trash"
HazLoot.LootColours = { "ffffff", "1eff00", "0080ff", "b048f8", "ff8000", "ff0000",}
HazLoot.LootColours[HazLoot.TRASH]   = "9d9d9d"

HazLoot.PhatLewt = {} -- Table of current distributable loot
HazLoot.CurrentItem  = nil  -- loot distro sequence number of loot item being distributed
HazLoot.LastWinner   = {}   -- winner of last mediated distro
HazLoot.SuicideOrder = {}   -- record of all "hard" loot suicided for, by sequence suicided
HazLoot.Suicides     = {}   -- record of all "hard" loot suicided for by name (see format below)
HazLoot.Offspecs     = {}   -- record of all "hard" loot rolled for   ({name={link,link,link,},})

HazLoot.STATE_IDLE           = 0
HazLoot.STATE_LOOTING        = 1
HazLoot.STATE_BID_OPEN       = 2
HazLoot.STATE_BID_LASTCALL   = 3
HazLoot.STATE_BID_CONFIRMED  = 4
HazLoot.STATE_ROLL_OPEN      = 5
HazLoot.STATE_ROLL_CONFIRMED = 6
HazLoot.STATE_SENDING_ITEM   = 7
HazLoot.State = HazLoot.STATE_IDLE

local options = {
  name    = "HazLoot",
  handler = HazLoot,
  type    = 'group',
  args    = {
    debug = {
      type = "toggle",
      name = "Debug",
      desc = "Enable debug messages in chat.",
      get  = "IsDebug",
      set  = "ToggleDebug",
    },
    useSKList = {
      type = "toggle",
      name = "SuicideKings",
      desc = "Use a Suicide Kings list for loot.",
      get  = "IsUsingSKList",
      set  = "ToggleUsingSKList",
    },
    softThreshold = {
      type = "select",
      name = "SoftThreshold",
      desc = "Loot above this quality, but below the hard threshold will be randomly and evenly passed out to party/raid members.",
      values = HazLoot.LootTypes, 
      get    = "GetSoftThreshold",
      set    = "SetSoftThreshold",
      style  = "radio",
    },
    hardThreshold = {
      type = "select",
      name = "HardThreshold",
      desc = "Loot above this quality will be kept for mediated assignment",
      values = HazLoot.LootTypes,
      get    = "GetHardThreshold",
      set    = "SetHardThreshold",
      style  = "radio",
    },
    reset = {
      type = "execute",
      name = "Reset Party List",
      desc = "Reset the party list, and clear random loot assignments",
      func = "ResetMemberList",
    },
    fullreset = {
      type = "execute",
      name = "Reset All Data",
      desc = "Reset the party list, clear random loot assignments, and clear the current loot list",
      func = "FullReset",
    },
    status = {
      type = "execute",
      name = "Status",
      desc = "Report who's received what",
      func = "StatusReport",
    },
    next = {
      type = "execute",
      name = "SuicideNext",
      desc = "Start suicide bidding for next item",
      func = "SuicideNext",
    },
    lastcall = {
      type = "execute",
      name = "SuicideLastCall",
      desc = "Last call for suicide bids on current item",
      func = "LastCall",
    },
    offspec = {
      type = "execute",
      name = "OffSpecRoll",
      desc = "Call for off-spec rolls for current item",
      func = "OffSpecRoll",
    },
    confirm = { -- This is a CYA announcement until the SK list code is written
      type  = "input",
      name  = "ConfirmWinner",
      desc  = "Confirm that you've read the SK list right",
      usage = "<Suicide Winner>",
      get   = "GetCurrentWinner",
      set   = "ConfirmWinner",
    },
    award = {
      type = "execute",
      name = "Award",
      desc = "Award loot to a winning player",
      func = "AwardCurrentItem",
    },
  },
}
-- Shortcuts
options.args.a  = options.args.award
options.args.c  = options.args.confirm
options.args.fr = options.args.fullreset
options.args.lc = options.args.lastcall
options.args.os = options.args.offspec

-- State
HazLoot.PartyMembers = {} -- int  -> name
HazLoot.PartyIndex   = {} -- name ->  int
HazLoot.LootList    = {}  -- TODO: persist guild-members in raids to DB
HazLoot.PartyCount = 0
HazLoot.LeastLoot  = {}
HazLoot.InRaid     = false
HazLoot.InParty    = true
HazLoot.SoftThreshold = HazLoot.UNCOMMON
HazLoot.HardThreshold = HazLoot.EPIC

-- State tracking for lazily loaded tables
HazLoot.skListLoaded    = false
HazLoot.partyListLoaded = false

local defaults = {
  profile = {
    debug         = true,
    useSKList     = true,
    softThreshold = HazLoot.UNCOMMON,
    hardThreshold = HazLoot.EPIC,
  },
}

------- ACE AND ADDON EVENTS ---------

function HazLoot:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("HazLootDB", defaults)
  if (nil == self.db) then
    self:Print("DB uninitialized... weird")
    self.db = {}
  else
    self:Print("Debug is "..(self.db.profile.debug and "ON" or "OFF"))
  end
  LibStub("AceConfig-3.0"):RegisterOptionsTable("HazLoot", options)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("HazLoot", "HazLoot")
  self:RegisterChatCommand("hazloot", "ChatCommand")
  self:RegisterChatCommand("hl", "ChatCommand")
end

function HazLoot:OnEnable()
  -- Called when the addon is enabled
  HazLoot:ResetMemberList(true) -- load member list, if there is one
  self:Print("I can haz loot!")

  self:RegisterEvent("LOOT_OPENED")
  self:RegisterEvent("LOOT_SLOT_CLEARED")
  self:RegisterEvent("PARTY_MEMBERS_CHANGED")
  self:RegisterEvent("RAID_ROSTER_UPDATE")
  self:RegisterEvent("UPDATE_INSTANCE_INFO")

  if self.db.profile.useSKList and not self.skListLoaded then
    self:LoadSKList()
  end
end

function HazLoot:OnDisable()
  -- Called when the addon is disabled
  HazLoot.skListLoaded    = false
  self:UnloadClassificationList()
  self:UnloadSKList()
end

--- WOW IN-GAME EVENTS ---

function HazLoot:UPDATE_INSTANCE_INFO()
  -- Check to see if this is a raid instance
  -- Check that you're NOT autolooting, if so, turn off autoloot
  -- Check for a duster - if no duster, warn every 30s (persist duster)
end

function HazLoot:PARTY_MEMBERS_CHANGED()
  self:GroupChange()
end

function HazLoot:RAID_ROSTER_UPDATE()
  self:GroupChange()
end

function HazLoot:LOOT_OPENED()
  HazLoot:LootThatBag()
end

function HazLoot:LOOT_SLOT_CLEARED(event, slot)
  HazLoot:LootCleared(slot)
end

-------- CORE ----------

function HazLoot:LootThatBag()
  if 0 == GetNumLootItems() then CloseLoot() return end -- Ignore empty bags
  if 0 < (#HazLoot.PhatLewt) then HazLoot.State = HazLoot.STATE_LOOTING end
  if UnitName("target") and UnitLevel("target") < 0 then
    HazLoot:Announce("Boss Loot! "..UnitName("target").." opened for HazLooting.")
  elseif not UnitName("target") then -- probably container (this may be wrong)
    HazLoot:Announce("Boss Loot! HazLooting.")
  end

  -- Iterate over the slots in the loot bag
  for slot = 1,GetNumLootItems() do
    if LootSlotIsCoin(slot) then
      LootSlot(slot)
    else
      -- This is kind of horrible, but seems necessary
      local itemLink   = GetLootSlotLink(slot)
      local itemString = HazLoot:GetItemString(itemLink)
      local id         = HazLoot:GetItemID(itemString)
      local name, link, rarity, ilevel, mlevel, itype, subtype, stack, loc, texture = GetItemInfo(id)
      if (rarity < self.db.profile.SoftThreshold) or "Money" == itype then -- Just loot it
        LootSlot(slot)
      elseif (rarity < self.db.profile.HardThreshold) then -- Random assign it
        HazLoot:AssignRandomLoot(rarity,name,link,slot,looterName)
      else -- itemRarity >= HazLoot.HardThreshold, so throw it in PhatLewt table
        table.insert(HazLoot.PhatLewt, {itemId = id, itemName = name, itemLink = link, loot_slot = slot })
        local c = (#HazLoot.PhatLewt)
      end
    end
  end

  if 0 < (#HazLoot.PhatLewt) then
    -- re-sort, because may have re-opened bag
    table.sort(HazLoot.PhatLewt, function(a,b) return HazLootList:LootOrderComparator(a,b) end)
    if HazLoot.State == HazLoot.STATE_IDLE then
      HazLoot:Announce("Distributing loot in this order: ")
      for i,loot in ipairs(HazLoot.PhatLewt) do
        -- self:Debug("finding roles for "..loot.itemName)
        loot.roles = HazLootList:GetRolesOf(loot.itemName)
        HazLoot:Announce(i..". "..loot.itemLink.." ("..loot.roles..")")
      end
      HazLoot:SuicideNext()
    end
  end
end

function HazLoot:SuicideNext()
  if not HazLoot.PhatLewt or 0 == (#HazLoot.PhatLewt) then
    self:Debug("Everything's distributed - or should be. Current item is "..(HazLoot.CurrentItem or "nil"))
    return
  end
  for order,loot in pairs(HazLoot.PhatLewt) do
    HazLoot.CurrentItem = order
    HazLoot:Warn("PST with suicides for: "..loot.itemLink.." ("..loot.roles..")")
    self:Debug(" /hl lc to issue the last call for "..loot.itemLink)
    HazLoot.State = HazLoot.STATE_BID_OPEN
    break
  end

  -- start a 15s timer1
  -- listen to chat for ("bid","suicide","me","want","yes",...)
    -- (later: conditional bidding)
  -- if whisper string does not contain bid, whisper usage
  -- if bidder can not equip, whisper warning
  -- for each bid, bump timer1 out by max(5s,timer1)
  -- when timer1 is up, announce "Last Call (10s)"
  -- start a 10s timer
  -- for each new bid, bump timer2 out by max(5s,timer2) and announce
  -- when timer2 is up:
    -- intersect bidder list with SK List (for order)
    -- if all have looted: winner is winner
    -- if any have not yet looted: winner is highest such
    -- set hasLooted bit on winner
    -- announce winner (and number of loots this raid)
    -- distribute loot
    -- reorder SK List
end

function HazLoot:LastCall()
  if HazLoot.State ~= HazLoot.STATE_BID_OPEN then
    self:Print("No open bids for last call.")
    return
  end
  HazLoot.State = HazLoot.STATE_BID_LASTCALL
  HazLoot:Warn("Last Call for "..HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink)
  self:Debug(" /hl confirm <name> to confirm the winner")
end

function HazLoot:OffSpecRoll()
  if HazLoot.State ~= HazLoot.STATE_BID_LASTCALL then
    self:Print("Must issue last call for suicides first: /hl lc")
    return
  end
  HazLoot.State = HazLoot.STATE_ROLL_OPEN
  HazLoot:Warn("Roll offspec for "..HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink)
  -- Run an off-spec roll for an item:
    -- start a 15s timer1
    -- listen to rolls
    -- if roller must suicide, whisper warning... and go to suicide?
    -- if roller can not equip, whisper warning
    -- for each bid, bump timer1 out by 5s
    -- when timer1 is up, announce "Off-spec rolls for <item> end in 10s"
    -- start a 10s timer
    -- for each new bid, bump timer2 out by max(5s,timer2) and announce
    -- when timer2 is up:
      -- announce winner
      -- distribute loot
end

function HazLoot:Dust(item)
  self:Debug("Distributing <item> for dusting.")
end

function HazLoot:ConfirmWinner(info,winner)
  if "" == winner and HazLoot.LastWinner.name then
    self:Debug("confirming "..HazLoot.LastWinner.name)
    if HazLoot.State == HazLoot.STATE_BID_LASTCALL then
      HazLoot.State = HazLoot.STATE_BID_CONFIRMED
    else --HazLoot.State ~= HazLoot.STATE_ROLL_OPEN
      HazLoot.State = HazLoot.STATE_ROLL_CONFIRMED
    end
    return
  end
  if (HazLoot.InRaid and not UnitInRaid(winner)) or not UnitInParty(winner) then
    self:Debug(winner.." doesn't seem to be in the raid/party. Check your spelling.")
    return
  end
  if HazLoot.State ~= HazLoot.STATE_BID_LASTCALL and HazLoot.State ~= HazLoot.STATE_ROLL_OPEN then
    self:Print("Issue a last call before confirming (/hl lc), or open for off-spec rolls (/hl os)")
    return
  end
  local lootCandidate = nil
  for i=1,40 do -- Loop over all possible master loot candidates
    self:Debug("DIE!"..i)
    lootCandidate = GetMasterLootCandidate(i) -- Use GMLC() here, because consistency doesn't matter: loot's been assigned
    if lootCandidate and winner == string.lower(lootCandidate) then
      winner = lootCandidate
      HazLoot.LastWinner.index  = i
      HazLoot.LastWinner.name   = winner
      HazLoot.LastWinner.link   = HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink
      HazLoot.LastWinner.slot   = HazLoot.PhatLewt[HazLoot.CurrentItem].loot_slot
      if HazLoot.State == HazLoot.STATE_BID_LASTCALL then
        HazLoot:Announce("Suiciding "..HazLoot.LastWinner.link.." to "..winner.." - speak now if you think I've got the SK List wrong, or file a ticket later")
      else --HazLoot.State ~= HazLoot.STATE_ROLL_OPEN
        HazLoot:Announce("Awarding "..HazLoot.LastWinner.link.." to "..winner.." - speak now if you think I've misread the offspec rolls")
      end
      return
    end
  end
end

function HazLoot:AwardCurrentItem(info)
  if HazLoot.State ~= HazLoot.STATE_BID_CONFIRMED and HazLoot.State ~= HazLoot.STATE_ROLL_CONFIRMED then
    self:Print("Confirm the winner with /hl confirm <name>")
    return
  elseif HazLoot.State ~= HazLoot.STATE_BID_CONFIRMED then
    HazLoot.LastWinner.wintable = HazLoot.Suicides
    HazLoot.LastWinner.order = HazLoot.SuicideOrder
  else --HazLoot.State ~= HazLoot.STATE_ROLL_CONFIRMED then
    HazLoot.LastWinner.wintable = HazLoot.Offspecs
  end
  if HazLoot.LastWinner.slot  then self:Debug("Slot: "..HazLoot.LastWinner.slot) end
  if HazLoot.LastWinner.index then self:Debug("Index: "..HazLoot.LastWinner.index) end
  GiveMasterLoot(HazLoot.LastWinner.slot,HazLoot.LastWinner.index)
  HazLoot.State = HazLoot.STATE_SENDING_ITEM
end

function HazLoot:LootCleared(slot)
  if not HazLoot.CurrentItem then return end
  for itemId,loot in pairs(HazLoot.PhatLewt) do
    if slot == HazLoot.PhatLewt[HazLoot.CurrentItem].loot_slot then
      local looter = HazLoot.LastWinner.name
      local link   = HazLoot.LastWinner.link
      HazLoot:Announce("Grats "..looter.."! "..link)
      --HazLoot:GAnnounce("Grats "..looter.."! "..link)
      if not HazLoot.LastWinner.wintable[looter] then HazLoot.LastWinner.wintable[looter] = {} end
      table.insert(HazLoot.LastWinner.wintable[looter],link)
      if HazLoot.LastWinner.order then table.insert(HazLoot.LastWinner.order,link) HazLoot.LastWinner.order = nil end
      HazLoot.PhatLewt[HazLoot.CurrentItem] = nil
      HazLoot.CurrentItem = nil
      if 0 < (#HazLoot.PhatLewt) then
        HazLoot.State = HazLoot.STATE_LOOTING
      else
        HazLoot.State = HazLoot.STATE_IDLE
      end
      return
    end
  end
  self:Debug("Something may be wrong: I know a loot slot got cleared, but the item sent wasn't in the PhatLewt table.")
end

function HazLoot:AssignRandomLoot(itemRarity,itemName,itemLink,slot,looterName)
  -- 1. Pick a random starting point in the list for items of this rarity
  -- 2. Cycle through through that list once until:
  -- 3a. EITHER find someone who has LeastLoot loots of that rarity, and award it
  -- 3b. OR find no nobody with LeastLoot lots of that rarity, award it to the random start and ++LeastLoot
  local tmpRand = math.random(40)
  local looterName
  for i = 1,40 do
    local randomLooter = (i+tmpRand)%40 + 1      -- +1 for LUA 1-based table indexes
    -- local randomLooter = math.mod(i+tmpRand,40) + 1      -- +1 for LUA 1-based table indexes
    looterName   = HazLoot.PartyMembers[randomLooter]
    if nil ~= looterName then
      -- self:Debug("i: "..i..", 1 + (("..i.."+"..tmpRand..")%"..HazLoot.PartyCount.."): "..randomLooter..", name: "..looterName)
      -- self:Debug("LeastLoot = "..HazLoot.LeastLoot[itemRarity]..", "..looterName.." has "..HazLoot.LootList[itemRarity][looterName].." loots of rarity "..HazLoot.LootTypes[itemRarity])
      if (HazLoot.LeastLoot[itemRarity] >= HazLoot.LootList[itemRarity][looterName]) then
        -- self:Debug("...which is the same as the least loot... so awarding loot to "..looterName)
        return HazLoot:GiveRandomLoot(itemRarity,itemName,itemLink,slot,looterName)
      end
    end
  end
  -- looterName = HazLoot.PartyMembers[math.mod(tmpRand,HazLoot.PartyCount) + 1]
  looterName = HazLoot.PartyMembers[tmpRand%HazLoot.PartyCount + 1]
  HazLoot.LeastLoot[itemRarity] = HazLoot.LeastLoot[itemRarity] + 1
  HazLoot:Announce("Everyone should have received "..HazLoot.LeastLoot[itemRarity].." "..HazLoot.LootTypes[itemRarity].." lewt(s). If you haven't please let Haz know.")
  return HazLoot:GiveRandomLoot(itemRarity,itemName,itemLink,slot,looterName)
end

function HazLoot:GiveRandomLoot(itemRarity,itemName,itemLink,slot,looterName)
  for i=1,40 do -- Might as well loop over all possible slots
    if looterName == GetMasterLootCandidate(i) then -- Use GMLC() here, because consistency doesn't matter: loot's been assigned
      GiveMasterLoot(slot,i)
      if nil == HazLoot.LootList[itemRarity][looterName] then
        HazLoot.LootList[itemRarity][looterName] = 1
      else
        HazLoot.LootList[itemRarity][looterName] = HazLoot.LootList[itemRarity][looterName] + 1
      end
      HazLoot:Announce(looterName.." got "..itemLink.." ("..HazLoot.LootList[itemRarity][looterName].." "..HazLoot.LootTypes[itemRarity].."(s) this session).")
      return true
    end
  end
  self:Debug("For some reason, "..looterName.." was not a valid master loot candidate.")
  return false
end

function HazLoot:StatusReport(info)
  local raider, tmpString
  for i=1,40 do
    raider = HazLoot.PartyMembers[i]
    if (nil ~= raider) then
      tmpString = ""
      for rarity = self.db.profile.SoftThreshold,self.db.profile.HardThreshold-1 do
        if nil ~= raider then
          tmpString = tmpString..HazLoot:RarityMessage(HazLoot.LootList[rarity][raider].." "..HazLoot.LootTypes[rarity]..", ",rarity)
        end
      end
      tmpString = tmpString..HazLoot:RarityMessage("to "..raider,0) -- trash colored!
      if HazLoot.Suicides[raider] then
        self:Debug("iteration over HazLoot.Suicides["..raider.."]")
        for i,win in ipairs(HazLoot.Suicides[raider]) do
          tmpString = tmpString.."\n\tSuicide: "..win
        end
      else tmpString = tmpString..HazLoot:RarityMessage(" (no suicides)",4)
      end
      if HazLoot.Offspecs[raider] then
        for i,win in ipairs(HazLoot.Offspecs[raider]) do
          tmpString = tmpString.."\n\tOffspec: "..win
        end
      else tmpString = tmpString..HazLoot:RarityMessage(" (no offspec)",4)
      end
      self:Debug(tmpString)
    end
  end
  self:Debug(HazLoot.PartyCount.." members in the raid")
end

function HazLoot:LoadClassificationList()
  self:Debug("loading classification list")
  -- Do I have a CL saved?
  -- If yes, load it.
  -- If no, load classifications.txt
end

function HazLoot:UnloadClassificationList()
  self:Debug("unloading classification list")
end

function HazLoot:LoadSKList()
  if self.db.profile.useSKList then
    self:Debug("loading SK list")
    self.skListLoaded = true
  else
    self:Debug("NOT loading SK List: it's disabled")
  end
end

function HazLoot:UnloadSKList()
  self:Debug("unloading SK list")
  self.skListLoaded = false
end

function HazLoot:GetCurrentWinner()
end

------- PARTY LIST HANDLING ------

--TODO: look into Ace-Bucket
function HazLoot:GroupChange()
  if HazLoot.partyListLoaded then
    HazLoot:UpdateMemberList()
  else
    HazLoot:ResetMemberList(true) -- reload
  end
end

function HazLoot:UpdateMemberList()
  local looter
  local tmpCurrentParty = {}
  tmpCurrentParty[UnitName("player")] = 0 -- Don't exclude yourself
  -- idempotent Add members still in the raid
  for i = 1,40 do
    looter = UnitName("raid"..i)
    if nil == looter then
      looter = UnitName("party"..i)
    end
    if (nil ~= looter) then
      tmpCurrentParty[looter] = i
      j = HazLoot:AddMemberToParty(looter)
      -- self:Debug("Found party member "..j..": "..looter)
    end
  end
  -- remove anybody who's no longer in the raid
  for i = 1,HazLoot.PartyCount do
    if nil == tmpCurrentParty[HazLoot.PartyMembers[i]] then
      HazLoot:UnloadPartyMember(i)
    end
  end
end

function HazLoot:UnloadPartyMember(index)
  -- TODO: make sure this is the correct behaviour with persistence
  if nil ~= HazLoot.PartyMembers[index] then
    HazLoot.PartyIndex[HazLoot.PartyMembers[index]] = nil
  end
  HazLoot.PartyMembers[index] = nil
  for rarity = self.db.profile.HardThreshold, self.db.profile.SoftThreshold, -1 do
    HazLoot.LootList[rarity][index] = nil
  end
  HazLoot.PartyCount = HazLoot.PartyCount - 1
end

function HazLoot:FullReset()
  HazLoot:ResetMemberList(true)
  HazLoot.PhatLewt = {} -- Table of current distributable loot
  HazLoot.CurrentItem  = nil
  HazLoot.LastWinner   = {}
end

function HazLoot:ResetMemberList(reload)
  for rarity = self.db.profile.HardThreshold, self.db.profile.SoftThreshold, -1 do
    HazLoot.LootList[rarity]  = nil
    HazLoot.LeastLoot[rarity] = 0
  end
  HazLoot.PartyMembers = nil
  HazLoot.PartyIndex   = nil
  HazLoot.PartyCount   = 0
  HazLoot.PartyIndex   = {}
  HazLoot.PartyMembers = {}
  for i=self.db.profile.SoftThreshold, self.db.profile.HardThreshold do
    HazLoot.LootList[i] = nil
    HazLoot.LootList[i] = {}
  end
  HazLoot.partyListLoaded = false

  if reload then
    HazLoot.partyListLoaded = HazLoot:LoadPartyMembers()
  end
end

function HazLoot:LoadPartyMembers()
  if (0 >= GetNumRaidMembers()) then
    self:Debug("Not in raid, getting party members")
    HazLoot.InParty = true
  else
    self:Debug("In Raid, getting raid members")
    HazLoot.InRaid = true
  end
  local looter, j
  for i = 1,40 do
    looter = UnitName(i < 6 and "party"..i or "raid"..i)
    if (nil ~= looter) then
      j = HazLoot:AddMemberToParty(looter)
      -- self:Debug("Found party member "..j..": "..looter)
    end
  end
  -- Don't forget yourself!
  looter = UnitName("player")
  j = HazLoot:AddMemberToParty(looter)
  -- self:Debug("Found party member "..j..": "..looter)
  self:Debug(HazLoot.PartyCount..(HazLoot.InRaid and " raiders loaded " or " party members loaded"))
  return true
end

-- This function should be idempotent
function HazLoot:AddMemberToParty(looter)
  local i = HazLoot.PartyIndex[looter]
  if nil ~= i then return i end -- They're already in the list. Move on.
  -- otherwise, iterate over the max raid table size, looking for the first empty slot
  for i = 1,40 do -- 40 is largest possible raid size, so we can't need more of a table than that
    if nil == HazLoot.PartyMembers[i] then
      HazLoot.PartyMembers[i]    = looter
      HazLoot.PartyIndex[looter] = i
      for rarity = self.db.profile.SoftThreshold,self.db.profile.HardThreshold-1 do
        HazLoot.LootList[rarity][looter] = HazLoot.LeastLoot[rarity]
      end
      HazLoot.PartyCount = HazLoot.PartyCount + 1
      return i
    elseif looter == HazLoot.PartyMembers[i] then -- this path should now be redundant
      self:Debug("ERROR: "..looter.." is already party member #"..i.." but I should already know that")
      return i
    end
  end
  self:Debug("I couldn't find room for "..looter.." have you got more than 40 in the raid, or am I broken?")
  return -1
end

------- UTILITY FUNCTIONS ------

function HazLoot:RarityMessage(message,rarity)
  if (nil == HazLoot.LootColours[rarity]) then return message end
  return "|cff"..HazLoot.LootColours[rarity]..message.."|r"
end

function HazLoot:GetItemString(itemLink)
  local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
  return itemString
end

function HazLoot:GetItemID(itemString)
  local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemString)
  return itemId
end

function HazLoot:Guild(message)
  SendChatMessage(message, "GUILD")
end

function HazLoot:Warn(message)
  SendChatMessage(message, "RAID_WARNING")
end

function HazLoot:GAnnounce(message)
  SendChatMessage(message, "GUILD")
end

function HazLoot:Announce(message)
  if HazLoot.InRaid then
    SendChatMessage(message, "RAID")
  elseif HazLoot.InParty then
    SendChatMessage(message, "PARTY")
  else
    self:Debug(message)
  end
end

-- Try to use tekDebug
-- local debugf = tekDebug and tekDebug:GetFrame("HazLoot")
function HazLoot:Debug(message)
  if self.db.profile.debug then
    --[[
    if debugf then
      debugf:AddMessage(string.join(",", ...))
    else
    ]]--
      self:Print(message)
    --end
  end
end

function HazLoot:ColouredRarity(rarity)
  return HazLoot:RarityMessage(HazLoot.LootTypes[rarity], rarity)
end
function HazLoot:PlainRarity(colouredRarity)
  -- "|cAARRGGBBRarity|r"
  -- "123456789012345678"
  return string.sub(colouredRarity,11,16)
end

------- GETTERS AND SETTERS FOR OPTIONS ---------

function HazLoot:IsDebug(info)
  return self.db.profile.debug
end
function HazLoot:ToggleDebug(info, newValue)
  self.db.profile.debug = newValue
end

function HazLoot:IsUsingSKList(info)
  return self.db.profile.useSKList
end
function HazLoot:ToggleUsingSKList(info, newValue)
  self.db.profile.useSKList = newValue
end

function HazLoot:GetHardThreshold(info)
  return self.db.profile.HardThreshold
end
function HazLoot:SetHardThreshold(info,newValue)
  self.db.profile.HardThreshold = newValue
  -- self:Debug("Hard Threshold now: "..HazLoot.LootTypes(newValue))
end

-- TODO: fix these so they show the nice coloured versions in the UI
function HazLoot:GetSoftThreshold(info)
  -- return HazLoot:ColouredRarity(self.db.profile.softThreshold)
  return self.db.profile.softThreshold
end
function HazLoot:SetSoftThreshold(info,newValue)
  --self.db.profile.softThreshold = HazLoot:PlainRarity(newValue)
  self.db.profile.softThreshold = newValue
end

function HazLoot:GetLastWinner(info)
  if HazLoot.LastWinner.name then
    self:Debug("Last winner was "..HazLoot.LastWinner.name..", who won "..HazLoot.LastWinner.link)
    return HazLoot.LastWinner.name
  else
    self:Debug("No loot awarded yet")
  end
  return ""
end

function HazLoot:ChatCommand(input)
  if not input or input:trim() == "" then
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
  else
    LibStub("AceConfigCmd-3.0").HandleCommand(HazLoot, "hl", "HazLoot", input)
  end
end

-------- TESTING STUFF --------

if HazLoot.standalone then
  self:Print("Standalone")
  --TODO: write some tests here!
end

